/* greg's ipaq alarm */
/* copyright (c) 2004 greg alexander (greg@galexander.org) */
/* $Id: gia.c,v 1.23 2004/04/14 10:05:24 galexand Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <time.h>
#include <linux/rtc.h>
#include <linux/apm_bios.h>
#include <sys/soundcard.h>
#include <ggl.h>

#define STATE_FILE "/tmp/gia_state"
#define LOG_FILE "/tmp/gia_log"
#define SOUND_FILE "/dev/dsp"

static void
logit(char *s,int x)
{
	static FILE *logit = NULL;
	if (!logit) logit = fopen("/tmp/gia_logit","a");
	if (!logit) { perror("/tmp/gia_logit"); return; }
	fprintf(logit,"%lu ",(unsigned long)time(NULL));
	fprintf(logit, s,x);
	fprintf(logit,"\n");
}

static void win_make(void);
static void state_transition(void);
static void state_load(void);
static void state_save(void);
static void log_wakeup(void);
static int update_time(void);
static void do_sleep(int howlong);
static void do_shutup(void);
static void sound_start(void);
static void sound_stop(void);
static void sound_doit(void);


/*  state |                display | button1 | button2
 * awake  | time since last wakeup | sleep   | quit
 * sleep  | time until wakeup      | sleep   | reset
 * wakeup | time since wakeup      | shutup  | shutup
 */
struct state_rec {
	enum {
		st_awake,
		st_sleep,
		st_wakeup
	} state;
	time_t last_wakeup;
	time_t last_sleep;
	time_t next_wakeup;
	time_t sleep_up;
	int awake_curr;
};

static struct state_rec state = { st_awake, 0, 0, 0, 60 };

static struct ggl_window *win;
static struct ggl_button *button1, *button2, *time_label;
struct ggl_gridbox *topgrid, *awakebar, *sleepbar, *wakeupbar;

/* awakebar: 10, 15, 20, 30, 45, 60, 90, 2H, 3H, 4H, 5H, 6H */
static struct ggl_button *awake_buttons[12];
static const int awake_amounts[12] = { 10, 15, 20, 30, 45, 60, 90, 120, 180, 240, 300, 360 };

/* wakeupbar: 0,1,2,3,4,5,6,7,8 */
static struct ggl_button *wakeup_buttons[9];
int wakeup_curr = 0;
int wakeup_ok = 0;

int
main(int argc, char **argv)
{
	srand(time(NULL));

	ggl_init();

	state_load();
	win_make();

	ggl_loop();

	return 0;
}

static void
win_destroy(struct ggl_widget *w)
{
        state_save();
        ggl_quit();
}

static void
button1_clicked(struct ggl_button *b)
{
	time_t t;
	time(&t);
	switch (state.state) {
		case st_awake:	/* sleep (for the default delay) */
			state.next_wakeup = t + (awake_amounts[state.awake_curr]*60);
			state.last_sleep = t;
			do_sleep(state.next_wakeup-t);
			break;
		case st_sleep:	/* sleep (for remainder on clock) */
			do_sleep(state.next_wakeup-t);
			break;
		case st_wakeup:	/* shut up */
			do_shutup();
			break;
	}
}

static void
button2_clicked(struct ggl_button *b)
{
	switch (state.state) {
		case st_awake:	/* quit */
			state_save();
			ggl_quit();
			break;
		case st_sleep:	/* reset */
			state.state = st_awake;
			state_transition();
			break;
		case st_wakeup:	/* shut up */
			do_shutup();
			break;
	}
}

static void
woke_up_clicked(struct ggl_button *b)
{
	time_t t;
	time(&t);
	state.state = st_awake;
	state.last_wakeup = t;
	log_wakeup();
	state_transition();
}

static void
delay_changed(struct ggl_button *b)
{
	int awake_new = state.awake_curr;
	struct ggl_button *o;
	int i;

	for (i = 0; i < 12; i++) {
		if (b == awake_buttons[i])
			awake_new = i;
	}
	o = awake_buttons[state.awake_curr];
	o->colors[0] = gc_yellow2;
	o->colors[1] = gc_white;
	o->parent.redraw((struct ggl_widget *)o);
	b->colors[0] = gc_blue3;
	b->colors[1] = gc_white;
	b->parent.redraw((struct ggl_widget *)b);
	state.awake_curr = awake_new;
	return;
}

/* sets up the wakeup buttons with a random one already selected,
 * and you have to press the one corresponding to (n*n)%9 to enable
 * the 'SHUTUP' buttons.
 * avoid_in is the number to avoid using again, generally used to avoid
 * repeating a number, allowing the user to try them sequentially.
 * avoid_out is the number to avoid having as an answer, generally
 * used to avoid the user just repeating a number until it becomes good.
 */
static void
wakeup_setup(int avoid_in, int avoid_out)
{
	int candidate = 0;
	struct ggl_button *o;

	do {
		candidate = rand() % 9;
	} while (candidate == avoid_in ||
	         ((candidate*candidate)%9) == avoid_out);
	wakeup_curr = candidate;
	o = wakeup_buttons[wakeup_curr];
	o->colors[0] = gc_blue3;
	o->colors[1] = gc_white;
	o->parent.redraw((struct ggl_widget *)o);
}

static void
try_wakeup(struct ggl_button *b)
{
	struct ggl_button *o;
	int wakeup_new = 0;
	int i;

	for (i = 0; i < 9; i++) {
		if (b == wakeup_buttons[i])
			wakeup_new = i;
	}
	o = wakeup_buttons[wakeup_curr];
	o->colors[0] = gc_yellow2;
	o->colors[1] = gc_yellow3;
	o->parent.redraw((struct ggl_widget *)o);
	if (((wakeup_curr * wakeup_curr) % 9) == wakeup_new) {
		wakeup_ok = 1;
		button1->colors[0] = gc_yellow2;
		button1->colors[1] = gc_yellow3;
		button1->parent.redraw((struct ggl_widget *)button1);
		button2->colors[0] = gc_yellow2;
		button2->colors[1] = gc_yellow3;
		button2->parent.redraw((struct ggl_widget *)button2);
	} else {
		wakeup_setup(wakeup_curr, wakeup_new);
	}
}

static void
state_transition(void)
{
	int i;

logit("state_transition(%d)\n",state.state);

	/* detach all of the possible old ones, to have just the one we want */
	ggl_gridbox_detach(topgrid, (struct ggl_widget *)awakebar);
	ggl_gridbox_detach(topgrid, (struct ggl_widget *)sleepbar);
	ggl_gridbox_detach(topgrid, (struct ggl_widget *)wakeupbar);

	switch (state.state) {
		case st_awake:
			button1->label = "SLEEP";
			button1->colors[0] = gc_yellow3;
			button1->colors[1] = gc_white;
			button2->label = "QUIT";
			button2->colors[0] = gc_yellow1;
			button2->colors[1] = gc_yellow2;
			if (state.awake_curr < 12) {
				awake_buttons[state.awake_curr]->colors[0] = gc_blue3;
				awake_buttons[state.awake_curr]->colors[1] = gc_white;
			}
			ggl_gridbox_attach(topgrid, (struct ggl_widget *)awakebar, "dd,ge");
			break;
		case st_sleep:
			button1->label = "SLEEP";
			button1->colors[0] = gc_yellow3;
			button1->colors[1] = gc_white;
			button2->label = "RESET";
			button2->colors[0] = gc_yellow1;
			button2->colors[1] = gc_yellow2;
			ggl_gridbox_attach(topgrid, (struct ggl_widget *)sleepbar, "dd,ge");
			break;
		case st_wakeup:
			wakeup_ok = 0;
			button1->label = "SHUTUP";
			button1->colors[0] = gc_blue1;
			button1->colors[1] = gc_blue1;
			button2->label = "SHUTUP";
			button2->colors[0] = gc_blue1;
			button2->colors[1] = gc_blue1;
			for (i = 0; i < 9; i++) {
				wakeup_buttons[i]->colors[0] = gc_yellow2;
				wakeup_buttons[i]->colors[1] = gc_yellow3;
			}
			ggl_gridbox_attach(topgrid, (struct ggl_widget *)wakeupbar, "dd,ge");
			break;
	}
	win->parent.resize((struct ggl_widget *)win,NULL,0);
	win->parent.redraw((struct ggl_widget *)win);
}

static void
win_make(void)
{
	struct ggl_button *b;
	struct ggl_gridbox *g;
	int i;

	win = ggl_window_new(320,200);
	win->parent.destroy = win_destroy;

	topgrid = ggl_gridbox_new("A10,B5,C5,d1,E5,f1,G5,H15,i0,z0","Y3,A20,b1,C5,D50,e1,F3");

	b = ggl_button_new(gbk_corner);
	b->parent.mouse = NULL;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "ac,bc,ca,cb");
	b = ggl_button_new(gbk_corner);
	b->parent.mouse = NULL;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "ad,bd,ce,cf");

	g = ggl_gridbox_new("a1,B30,z0","a1,z0");
	b = ggl_button_new(gbk_square);
	b->parent.mouse = NULL;
	b->colors[0] = gc_blue2;
	ggl_gridbox_attach(g, (struct ggl_widget *)b, "aa,bz");
	b = ggl_button_new(gbk_black);
	b->parent.mouse = NULL;
	b->label = "GIA";
	b->font = gf_large;
	ggl_gridbox_attach(g, (struct ggl_widget *)b, "ba,zz");
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)g, "da,gb");

	b = ggl_button_new(gbk_square);
	b->u.square.rounded[1] = 1.0;
	b->u.square.rounded[2] = 1.0;
	b->colors[0] = gc_blue1;
	b->parent.mouse = NULL;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "ha,ib");

	b = ggl_button_new(gbk_black);
	b->parent.mouse = NULL;
	b->font = gf_largest;
	b->label_pos = 4;
	b->colors[0] = gc_orange3;
	b->colors[1] = gc_blue3;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "db,gc");
	time_label = b;

	awakebar = ggl_gridbox_new("a1,B5,c1,D5,e1,F5,g1,H5,i1,J5,k1,z0","a1,B5,c1,D9,z0");
	for (i = 0; i < 12; i++) {
		awake_buttons[i] = ggl_button_new(gbk_square);
		awake_buttons[i]->colors[0] = gc_yellow2;
		awake_buttons[i]->colors[1] = gc_yellow3;
		awake_buttons[i]->clicked = delay_changed;
	}
	awake_buttons[0]->label = "10";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[0],"aa,bb");
	awake_buttons[1]->label = "15";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[1],"ca,db");
	awake_buttons[2]->label = "20";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[2],"ea,fb");
	awake_buttons[3]->label = "30";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[3],"ga,hb");
	awake_buttons[4]->label = "45";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[4],"ia,jb");
	awake_buttons[5]->label = "60";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[5],"ka,zb");
	awake_buttons[6]->label = "90";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[6],"ac,bd");
	awake_buttons[7]->label = "2H";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[7],"cc,dd");
	awake_buttons[8]->label = "3H";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[8],"ec,fd");
	awake_buttons[9]->label = "4H";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[9],"gc,hd");
	awake_buttons[10]->label = "5H";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[10],"ic,jd");
	awake_buttons[11]->label = "6H";
	ggl_gridbox_attach(awakebar, (struct ggl_widget *)awake_buttons[11],"kc,zd");

	sleepbar = ggl_gridbox_new("A5,b1,C5,z0","A10,b1,C10,z0");
	b = ggl_button_new(gbk_square);
	b->u.square.rounded[0] = 1.0; b->u.square.rounded[1] = 1.0;
	b->u.square.rounded[2] = 1.0; b->u.square.rounded[3] = 1.0;
	b->label = "WOKE UP";
	b->font = gf_large;
	b->clicked = woke_up_clicked;
	ggl_gridbox_attach(sleepbar, (struct ggl_widget *)b, "bb,cc");

	wakeupbar = ggl_gridbox_new("a1,B3,c1,D3,e1,F3,g1,H3,i1,J3,k1,L3,m1,N3,o1,P3,q1,z0","A10,b1,C10,z0");
	for (i = 0; i < 9; i++) {
		wakeup_buttons[i] = ggl_button_new(gbk_square);
		wakeup_buttons[i]->colors[0] = gc_yellow2;
		wakeup_buttons[i]->colors[1] = gc_yellow3;
		wakeup_buttons[i]->clicked = try_wakeup;
	}
	wakeup_buttons[0]->label = "0";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[0],"ab,bc");
	wakeup_buttons[1]->label = "1";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[1],"cb,dc");
	wakeup_buttons[2]->label = "2";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[2],"eb,fc");
	wakeup_buttons[3]->label = "3";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[3],"gb,hc");
	wakeup_buttons[4]->label = "4";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[4],"ib,jc");
	wakeup_buttons[5]->label = "5";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[5],"kb,lc");
	wakeup_buttons[6]->label = "6";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[6],"mb,nc");
	wakeup_buttons[7]->label = "7";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[7],"ob,pc");
	wakeup_buttons[8]->label = "8";
	ggl_gridbox_attach(wakeupbar, (struct ggl_widget *)wakeup_buttons[8],"qb,zc");

	b = ggl_button_new(gbk_square);
	b->clicked = button1_clicked;
	b->label_pos = 3;
	b->font = gf_larger;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "de,ef");
	button1 = b;
	b = ggl_button_new(gbk_square);
	b->clicked = button2_clicked;
	b->label_pos = 3;
	b->font = gf_larger;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "fe,gf");
	button2 = b;
	b = ggl_button_new(gbk_square);
	b->u.square.rounded[1] = 1.0;
	b->u.square.rounded[2] = 1.0;
	b->colors[0] = gc_blue1;
	b->parent.mouse = NULL;
	ggl_gridbox_attach(topgrid, (struct ggl_widget *)b, "he,if");

	ggl_window_attach(win, (struct ggl_widget *)topgrid);

	ggl_register_timer(1000, update_time);
	update_time();

	state_transition();
}

static void
state_load(void)
{
	FILE *f = fopen(STATE_FILE,"r");
	char buf[100];
	if (!f) {
		perror(STATE_FILE);
		return;
	}
	fgets(buf,100,f);
	state.state = strtoul(buf,NULL,10);
	fgets(buf,100,f);
	state.last_wakeup = strtoul(buf,NULL,10);
	fgets(buf,100,f);
	state.last_sleep = strtoul(buf,NULL,10);
	fgets(buf,100,f);
	state.next_wakeup = strtoul(buf,NULL,10);
	fgets(buf,100,f);
	state.awake_curr = strtoul(buf,NULL,10);
	fclose(f);
	state.sleep_up = 0;
}

static void
state_save(void)
{
	FILE *f = fopen(STATE_FILE,"w");
	if (!f) {
		perror(STATE_FILE);
		return;
	}
	fprintf(f,"%d\n",(int)state.state);
	fprintf(f,"%lu\n",(unsigned long)state.last_wakeup);
	fprintf(f,"%lu\n",(unsigned long)state.last_sleep);
	fprintf(f,"%lu\n",(unsigned long)state.next_wakeup);
	fprintf(f,"%d\n",(int)state.awake_curr);
	fclose(f);
}

static void
log_wakeup(void)
{
	FILE *f = fopen(LOG_FILE,"a");
	struct tm *tm;

	tm = localtime(&state.last_sleep);
	fprintf(f,"%02d-%02d-%02d %02d:%02d:%02d - ", tm->tm_year%100, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
	tm = localtime(&state.last_wakeup);
	fprintf(f,"%02d-%02d-%02d %02d:%02d:%02d (%d)\n", tm->tm_year%100, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, (int)(state.last_wakeup-state.last_sleep)/60);
	fclose(f);
}

static int
update_time(void)
{
	time_t t;
	static char buf[100] = "";
	time(&t);
logit("update_time(%d)",(int)state.state);
	switch (state.state) {
		case st_awake:
			t = t-state.last_wakeup;
			break;
		case st_sleep:
			if (state.sleep_up && state.sleep_up+60 <= t) {
				do_sleep(state.next_wakeup-t);
				time(&t);
			}
			if (!state.sleep_up) {
				state.sleep_up = t;
			}
			if (state.next_wakeup <= t) {
				state.state = st_wakeup;
				time_label->active = 1;
				state.last_wakeup = t;
				log_wakeup();
				state_transition();
				wakeup_setup(0,1);
				t = 0;
				sound_start();
			} else {
				t = state.next_wakeup-t;
			}
			break;
		case st_wakeup:
			t = t-state.next_wakeup;
			sound_doit();
			break;
	}

	buf[0] = 0;
	if (t >= 86400) {
		sprintf(buf+strlen(buf),"%dd ",(int)(t / 86400));
		t %= 86400;
	}
	sprintf(buf+strlen(buf),"%d ",(int)(t/3600));
	t %= 3600;
	sprintf(buf+strlen(buf),"%02d ",(int)(t/60));
	t %= 60;
	sprintf(buf+strlen(buf),"%02d",(int)(t));
	time_label->label = buf;
	time_label->parent.redraw((struct ggl_widget *)time_label);

	return 1;
}

static int
apm_suspend(void)
{
	int fd = open("/dev/apm_bios",O_RDWR);
	if (fd < 0) {
		perror("/dev/rtc");
		return -1;
	}
	if (ioctl(fd, APM_IOC_SUSPEND, 0) == -1) {
		perror("APM_IOC_SUSPEND");
		close(fd);
		return -1;
	}
	close(fd);

	sleep(1);	/* wait for clock to catch up */
	sleep(1);

	return 0;
}

static void
do_sleep(int howlong)
{
	int fd;
	struct tm tm;
	struct tm *tmp;
	time_t t;

logit("do_sleep(%d)",howlong);

	state.state = st_sleep;
	state.sleep_up = 0;
	state_transition();
	state.sleep_up = 0;

	if (howlong < 30) return;

	fd = open("/dev/rtc",O_RDWR);
	if (fd < 0) {
		perror("/dev/rtc");
		return;
	}
	if (ioctl(fd, RTC_RD_TIME, &tm) == -1) {
		perror("RTC_RD_TIME");
		close(fd);
		return;
	}
	t = mktime(&tm)+howlong;
	tmp = localtime(&t);

	if (ioctl(fd, RTC_ALM_SET, tmp) == -1) {
		perror("RTC_ALM_SET");
		close(fd);
		return;
	}
	if (ioctl(fd, RTC_AIE_ON, 0) == -1) {
		perror("RTC_AIE_ON");
		close(fd);
		return;
	}
logit("suspending",0);
	apm_suspend();
logit("resumed!",0);

	if (ioctl(fd, RTC_AIE_OFF, 0) == -1) {
		perror("RTC_AIE_OFF");
		close(fd);
		return;
	}

	close(fd);
}

static int sound_fd = -1;
static int sound_volume;

static void
do_shutup(void)
{
	if (wakeup_ok) {
		state.state = st_awake;
		state_transition();
		time_label->active = 0;
		sound_stop();
	} else if (sound_fd != -1) {
		int t = 0x3232;
		if (ioctl(sound_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &t) == -1)
			perror("MIXER_WRITE");
	}
}

static void
sound_start(void)
{
	int t;

	if (sound_fd != -1) return;		/* already open?? */

	sound_fd = open(SOUND_FILE,O_RDWR);
	if (sound_fd == -1) {
		perror(SOUND_FILE);
		return;
	}

	if (ioctl(sound_fd, MIXER_READ(SOUND_MIXER_VOLUME), &sound_volume) == -1)
		perror("MIXER_READ");

	t = 0x6464;
	if (ioctl(sound_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &t) == -1)
		perror("MIXER_WRITE");

	t = AFMT_S16_LE;		/* PORTABILITY -> _BE for big endian */
	if (ioctl(sound_fd,SNDCTL_DSP_SETFMT,&t) == -1 || t != AFMT_S16_LE)
		perror("ioctl(SNDCTL_DSP_SETFMT,AFMT_S16_LE)");
	t = 1;
	if (ioctl(sound_fd,SNDCTL_DSP_STEREO,&t) == -1 || t != 1)
		perror("ioctl(SNDCTL_DSP_STEREO,1)");

	t = 11025;
	if (ioctl(sound_fd,SNDCTL_DSP_SPEED,&t) == -1 || t < 10000 || t > 12000)
		perror("ioctl(SNDCTL_DSP_SPEED,11025)");
	t = 0x00041000;
	if (ioctl(sound_fd,SNDCTL_DSP_SETFRAGMENT,&t) == -1 || t != 0x00041000)
		perror("ioctl(SNDCTL_DSP_SETFRAGMENT,0x00041000)");
}

static void
sound_stop(void)
{
	int t = 0;
	if (sound_fd == -1) return;	/* already closed? */

	if (ioctl(sound_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &t) == -1)
		perror("MIXER_WRITE");
	close(sound_fd);
	if (ioctl(sound_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &sound_volume) == -1)
		perror("MIXER_WRITE");
	sound_fd = -1;
}

static void
sound_doit(void)
{
	static signed short buf[22050];
	static int buf_inited = 0;

	if (sound_fd == -1) return;	/* closed?? */

	if (!buf_inited) {
		int i;
		for (i = 0; i < 22050; i++) {
			buf[i] = rand();
		}
		buf_inited = 1;
	}

	write(sound_fd, buf, 11025*2*2);
}
