#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#define XK_MISCELLANY 1
#include <X11/keysymdef.h>
#include <ggl.h>
#include "gread.h"

static struct ggl_gridbox *readbox;
static struct ggl_button *label_file, *label_done;
static struct ggl_textlist *text;
static char *mmap_text = NULL;
static int mmap_fd = -1;
static int mmap_len = 0;


static int reader_keys(unsigned code, unsigned state, int down);
static void reader_up(struct ggl_button *b);
static void reader_down(struct ggl_button *b);
static void reader_bigup(struct ggl_button *b);
static void reader_bigdown(struct ggl_button *b);
static void reader_quit(struct ggl_button *b);
static void reader_load_text(struct ggl_textlist *t);
static void update_done(void);
static void close_mmap(void);
static void line_up(void);
static char *line_down(void);
static void prev_word(void);
static void next_word(void);


struct ggl_widget *
reader_get(void)
{
	struct ggl_button *b;

	if (!readbox) {
		readbox = ggl_gridbox_new("A25,B5,C3,d1,E3,F6,g0","A10,B3,C30,D3,e2,F3,g3,H3,i3,J3,k2,L3,m2,n0");

		b = ggl_button_new(gbk_corner);
		b->label_pos = 2;
		b->colors[0] = gc_purple1;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "ad,bd,ca,cb");
		label_done = b;
		b = ggl_button_new(gbk_square);
		b->colors[0] = gc_purple1;
		b->font = gf_smaller;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "da,eb");
		label_file = b;
		b = ggl_button_new(gbk_square);
		b->colors[0] = gc_peach;
		b->u.square.rounded[1] = 1.0;
		b->u.square.rounded[2] = 1.0;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "fa,gb");

		b = ggl_button_new(gbk_square);
		b->label = "UP";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = reader_bigup;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "ae,bf");
		b = ggl_button_new(gbk_square);
		b->label = "UP";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = reader_up;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "ag,bh");
		b = ggl_button_new(gbk_square);
		b->label = "DN";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = reader_down;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "ai,bj");
		b = ggl_button_new(gbk_square);
		b->label = "DN";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = reader_bigdown;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "ak,bl");
		b = ggl_button_new(gbk_square);
		b->label = "QUIT";
		b->label_pos = 2;
		b->colors[0] = gc_peach;
		b->colors[1] = gc_white;
		b->clicked = reader_quit;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)b, "am,bn");

		text = ggl_textlist_new();
		text->font = gf_small;
		text->colors[0] = gc_green;
		text->parent.mouse = NULL;
		text->load_text = reader_load_text;
		ggl_gridbox_attach(readbox, (struct ggl_widget *)text, "cc,fn");
	}

	if (file_states) {
		static char buf[100];
		strncpy(buf,file_states->fn,99);
		buf[99] = 0;
		label_file->label = buf;
	}
	update_done();

	ggl_register_keys(reader_keys);

	return (struct ggl_widget *)readbox;
}

static int
reader_keys(unsigned code, unsigned state, int down)
{
		/* only respond to releases */
	if (down) return 0;

	switch (code) {
		case XK_Up: reader_up(NULL); break;
		case XK_Right: reader_bigdown(NULL); break;
		case XK_Down: reader_down(NULL); break;
		case XK_Left: reader_bigup(NULL); break;
		default: return 0;
	}

	return 1;
}

static void
reader_up(struct ggl_button *b)
{
	int i;
	for (i = 0; i < text->num_lines-1; i++) {
		line_up();
	}

	text->parent.resize((struct ggl_widget *)text, NULL, 0);
	text->parent.redraw((struct ggl_widget *)text);
	update_done();
}

static void
reader_down(struct ggl_button *b)
{
	int i;
	for (i = 0; i < text->num_lines-1; i++) {
		line_down();
	}

	text->parent.resize((struct ggl_widget *)text, NULL, 0);
	text->parent.redraw((struct ggl_widget *)text);
	update_done();
}

static void
reader_bigup(struct ggl_button *b)
{
	int step = mmap_len / NUM_BIG_STEPS;

	if (file_states->offset > step) {
		file_states->offset -= step;
	} else {
		file_states->offset = 0;
	}
	
	text->parent.resize((struct ggl_widget *)text, NULL, 0);
	text->parent.redraw((struct ggl_widget *)text);
	update_done();
}

static void
reader_bigdown(struct ggl_button *b)
{
	int step = mmap_len / NUM_BIG_STEPS;
	int max = mmap_len - 200;

	file_states->offset += step;
	if (file_states->offset > max)
		file_states->offset = max;
	
	text->parent.resize((struct ggl_widget *)text, NULL, 0);
	text->parent.redraw((struct ggl_widget *)text);
	update_done();
}

static void
reader_quit(struct ggl_button *b)
{
	close_mmap();
	do_filelist();
}

static void
reader_load_text(struct ggl_textlist *t)
{
	unsigned long old_offset = file_states->offset;
	int i;

	for (i = 0; i < t->num_lines; i++) {
		t->lines[i] = safe_strdup(line_down());
	}
	file_states->offset = old_offset;
}

static void
update_done(void)
{
	if (file_states) {
		static char buf[100];
		sprintf(buf, "%03d",(int)((file_states->offset*100) / (file_states->total_len ? file_states->total_len : 1)));
		label_done->label = buf;
	} else {
		label_done->label = NULL;
	}
	label_done->parent.redraw((struct ggl_widget *)label_done);
}

int
reader_load(char *fn)
{
	close_mmap();

	if ((mmap_fd = open(fn,O_RDONLY)) == -1) {
		perror(fn);
		return -1;
	}
	mmap_len = lseek(mmap_fd, 0, SEEK_END);
	lseek(mmap_fd, 0, SEEK_SET);
	if (!(mmap_text = mmap(NULL, mmap_len, PROT_READ, MAP_SHARED, mmap_fd, 0))) {
		perror("mmap");
		close(mmap_fd);
		return -1;
	}

	file_states->total_len = mmap_len;

	return 0;
}

static void
close_mmap(void)
{
	if (mmap_fd != -1) {
		munmap(mmap_text, mmap_len);
		close(mmap_fd);
		mmap_text = NULL;
		mmap_fd = -1;
		mmap_len = 0;
	}
}

static void
line_up(void)
{
	static char *buf = NULL;
	static int buf_len = 0;
	unsigned long start_ofs = file_states->offset;
	unsigned long word_ofs;
	int line_len;
	char c;

	if (file_states->offset != 0) file_states->offset--;

	while (1) {
		word_ofs = file_states->offset;
		prev_word();
		line_len = start_ofs - file_states->offset;
		if (line_len+2 >= buf_len) {
			buf_len = line_len + 256;
			free(buf);
			buf = safe_malloc(buf_len);
		}
		memcpy(buf, mmap_text + file_states->offset, line_len);
		buf[line_len] = 0;
		if (ggl_textlist_percent(text, buf) > 100) {
			file_states->offset = word_ofs + 1;
			return;
		}
		if (file_states->offset == 0) {
			return;
		}
		file_states->offset--;
		c = mmap_text[file_states->offset];
		if (c == '\n') {
			file_states->offset++;
			return;
		}
	}
}

static char *
line_down(void)
{
	static char *buf = NULL;
	static int buf_len = 0;
	unsigned long start_ofs = file_states->offset;
	unsigned long word_ofs;
	int line_len;
	char c;

	while (1) {
		word_ofs = file_states->offset;
		next_word();
		line_len = file_states->offset - start_ofs;
		if (line_len+2 >= buf_len) {
			buf_len = line_len + 256;
			free(buf);
			buf = safe_malloc(buf_len);
		}
		memcpy(buf, mmap_text + start_ofs, line_len);
		buf[line_len] = 0;
		if (ggl_textlist_percent(text, buf) > 100) {
			file_states->offset = word_ofs;
			line_len = file_states->offset - start_ofs;
			memcpy(buf, mmap_text + start_ofs, line_len);
			buf[line_len] = 0;
			return buf;
		}
		if (file_states->offset >= mmap_len) {
			return buf;
		}
		c = mmap_text[file_states->offset];
		if (c == '\n') {
			file_states->offset++;
			return buf;
		}
		if (isspace(c)) {
			file_states->offset++;
		}
	}
}

static void
prev_word(void)
{
	char c;
	int i = 0;

	if (file_states->offset == 0)
		return;
	if (isspace(mmap_text[file_states->offset]))
		file_states->offset--;

	while (1) {
		if (file_states->offset == 0)
			return;
		c = mmap_text[file_states->offset];
		if (isspace(c)) {
			file_states->offset++;
			return;
		}
		file_states->offset--;
		i++;
		if (i >= MAX_WORD_LEN)
			return;
	}
}

static void
next_word(void)
{
	char c;
	int i = 0;

	while (1) {
		if (file_states->offset >= mmap_len)
			return;
		c = mmap_text[file_states->offset];
		if (isspace(c))
			return;
		file_states->offset++;
		i++;
		if (i >= MAX_WORD_LEN)
			return;
	}
}
